#define ESP8266
#include <Wire.h>
#include <EEPROM.h>
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include "SFE_BMP180.h"
#include "TM1638.h"
#include <stdio.h>

// Распиновка платы "Вемос" (Wemos)
#define D0                      16
#define D1                      5
#define D2                      4
#define D3                      0
#define D4                      2
#define D5                      14
#define D6                      12
#define D7                      13
#define D8                      15

// Режимы индикации
#define IND_VOL                 1
#define IND_DVOL                2
#define IND_LEV                 4
#define IND_P1                  8
#define IND_P2                  16
#define IND_T1                  32
#define IND_T2                  64
#define MODE_SET_D              128

// Подключение индикатора
#define TM1638_DIO              D7
#define TM1638_STB              D6
#define TM1638_CLK              D5

// Подключение датчиков давления
#define BMP180_1_SDA            D4
#define BMP180_1_SCL            D3
#define BMP180_2_SDA            D1
#define BMP180_2_SCL            D2

// Параметры для измерения уровня
#define P                       1000      // Плотность жидкости кг/м3
#define G                       9.8       // Ускорение свободного падения
//#define PI                    3.14      // Число пи
//#define D                     0.246     // Диаметр сосуда
#define N                       5         // Количество точек усреднения

#define AP_SSID                  "ESP_LEVEL"
#define AP_PASS                  "00000000000"

#define UDP_PORT                  8000

#define MODE_SERVER               1
#define MODE_CLIENT               2
#define MODE_ERROR                0


TM1638        tm1638_disp(TM1638_DIO, TM1638_CLK, TM1638_STB); //инициализация дисплея
SFE_BMP180    bmp180_1, bmp180_2;  // переменные для доступа к BMP180

WiFiServer http_server(80);
WiFiServer telnet_server(23);
WiFiClient client_telnet;
WiFiUDP udp;
  
struct data
{
  int int_10V;
  int int_10dV;
  int int_h;
  int int_P1;
  int int_P2;
  int int_T1;
  int int_T2;
} serv_data;

double d, V0 = 0;  
char d0, d1, d2, d3;
char Mode = IND_VOL;                        // Режим индикации
char ip_mode = MODE_ERROR;                 // тип устройства 





void setup() 
{
  tm1638_disp.clearDisplay();              // Очистка семисегментного индикатора, команда не влияет на светодиоды
  tm1638_disp.setLEDs(Mode);
  Serial.begin(115200); Serial.println("\n\n\n\nESP started OK!");

// Инициализация датчиков давления, определение типа устройства  

  Wire.begin(BMP180_1_SDA, BMP180_1_SCL);
  char temp1 = bmp180_1.begin();
  Wire.begin(BMP180_2_SDA, BMP180_2_SCL);
  char temp2 = bmp180_2.begin();

  if (temp1 && temp2)   // оба инициализировались нормально - сервер
  {
    Serial.println("BMP180 # 1 started OK");
    Serial.println("BMP180 # 2 started OK");
    Serial.println("\nMOODE = SERVER\n");
    ip_mode = MODE_SERVER;
  }
  else
  {
    if (temp1 == 0) Serial.println("BMP180 # 1 started ERROR!!!");
    if (temp2 == 0) Serial.println("BMP180 # 2 started ERROR!!!");
    if ((temp1 || temp2) == 0) 
    {
      ip_mode = MODE_CLIENT;                 // оба инициализировались ненормально - клиент
      Serial.println("\nMOODE = CLIENT\n");
    }
  }

  if (ip_mode == MODE_ERROR) { Serial.println("\nInit error. Reset ESP8266 ...\n"); ESP.reset(); }

  if (ip_mode == MODE_CLIENT)   client_setup(); // если это клиент
  else                          server_setup(); // если это сервер

  if ( udp.begin(UDP_PORT) )   Serial.println("UDP started OK");  // Иниализация UDP
 
  delay(1000);
}


void client_setup() 
{
  if( !ConnectWiFi(AP_SSID,AP_PASS) ) 
  { 
    Serial.println("No Wi-Fi connect. Reset ESP8266 ..."); 
    ESP.reset(); 
  }

  Serial.print("\n\nSSID: ");           Serial.println(WiFi.SSID());
  Serial.print("Signal power (RSSI):"); Serial.print(WiFi.RSSI()); Serial.println(" dBm");
  Serial.print("IP Address: ");         Serial.println(WiFi.localIP());
  Serial.print("Network mask: ");       Serial.println(WiFi.subnetMask());
  Serial.print("Network gateway: ");    Serial.println(WiFi.gatewayIP());
  Serial.println();
}

void server_setup() 
{
  EEPROM.begin(4);
  d0 = EEPROM.read(0); if (d0 > 9) d0=0;
  d1 = EEPROM.read(1); if (d1 > 9) d1=0;
  d2 = EEPROM.read(2); if (d2 > 9) d2=0;
  d3 = EEPROM.read(3); if (d3 > 9) d3=0;
  d = d3 + 0.1*d2 + 0.01*d1 + 0.001*d0;
  
  WiFi.mode(WIFI_AP);
  WiFi.softAP(AP_SSID, AP_PASS);
  Serial.print("AP IP: "); Serial.println(WiFi.softAPIP());   WiFi.printDiag(Serial);  Serial.println(); 
  
  http_server.begin();
  telnet_server.begin();
  
  Serial.println("HTTP, Telnet servers started OK");
}





void loop() 
{
  char temp = tm1638_disp.getButtons();
  
  if (temp) 
  {
    Mode = temp;
    tm1638_disp.setLEDs(Mode);
  }
    
  if (ip_mode == MODE_CLIENT)     // если клиент
  {
    do_client();
    do_led_ind();
    return;
  }


// если сервер
  
  double T1, T2, P1, P2, h, V, _P1, _P2, _T1, _T2;

  if (temp >= MODE_SET_D)    // если нажата кнопка установки
  {
    char set_mode = Mode - MODE_SET_D;
    
    if (set_mode == IND_DVOL) // если комбинация кнопок для сброса счетчика расхода
    {
      V0 = V;
      Mode = IND_DVOL;
      tm1638_disp.setLEDs(Mode);
      return;
    }
  
    if (set_mode == 0)  // если нажата только кнопка установки - отображаем заданный диаметр
    {
      tm1638_disp.setDisplayToSignedDecNumber( (int)(1000*d), 0, false); 
      return;
    }
    
  // иначе если нажата кнопка установки с кнопками изменения диаметра - отображаем диаметр
    d0 = (int(d*1000)) % 10;  
    d1 = (int(d*100)) % 10;   
    d2 = (int(d*10)) % 10;    
    d3 = (int(d)) % 10;       
    if (set_mode == 64) d0 += 1; if (d0 > 9) d0 = 0;
    if (set_mode == 32) d1 += 1; if (d1 > 9) d1 = 0;
    if (set_mode == 16) d2 += 1; if (d2 > 9) d2 = 0;
    if (set_mode == 8) d3 += 1;  if (d3 > 9) d3 = 0;
    
    d = d3 + 0.1*d2 + 0.01*d1 + 0.001*d0;
    tm1638_disp.setDisplayToSignedDecNumber( (int)(1000*d), 0, false);
  
    EEPROM.write(0, d0);
    EEPROM.write(1, d1);
    EEPROM.write(2, d2);
    EEPROM.write(3, d3);
    EEPROM.commit();
  
    delay(500);
    return;
  }

// Основные вычисления

  _P1 = 0; _P2 = 0; _T1 = 0; _T2 = 0; V = 0; h = 0;
  for (char i=0; i<N; i++)
  {
    readPressure(&T1, &T2, &P1, &P2);
    _P1 += P1;
    _P2 += P2;
    _T1 += T1;
    _T2 += T2;
  }
  P1 = _P1/N; P2 = _P2/N; T1 = _T1/N; T2 = _T2/N; 

  h = 1000*(P2-P1) / (P * G); // в мм
  V = PI*d*d/4 * h;

  if (h<0) h = 0;
  if (V<0) V = 0;

  if ( (V0 == 0) || (V>V0) ) V0 = V;

  serv_data.int_10V = (int)(10*V);          // литры с точностью 0,1
  serv_data.int_10dV = (int)(10*(V0-V));    // литры с точностью 0,1
  serv_data.int_h = (int)h;                 // мм
  serv_data.int_P1 = (int)P1;               // Па
  serv_data.int_P2 = (int)P2;               // Па
  serv_data.int_T1 = (int)(T1*10);          // *С с точностью 0,1
  serv_data.int_T2 = (int)(T2*10);          // *С с точностью 0,1

  do_led_ind();
  do_server();
}










void do_client()
{
// UDP
  char packetBuffer[sizeof(serv_data)];

  if (WiFi.status() != WL_CONNECTED) ESP.reset(); 
  
  if (udp.parsePacket()) 
  {
    udp.read(packetBuffer, sizeof(packetBuffer));

    if ( udp.remoteIP() == WiFi.gatewayIP() ) memcpy(&serv_data, packetBuffer, sizeof(serv_data));

    Serial.print(packetBuffer);
    
  }
/*
    udp.beginPacket(WiFi.gatewayIP(), UDP_PORT);
    udp.write("get");
    udp.endPacket();
*/
}



void do_server()
{
  WiFiClient client;
  char str[128];

// TELNET
  if (!client_telnet || !client_telnet.connected()) client_telnet = telnet_server.available();
  if (client_telnet && client_telnet.connected())  // пока сохраняется подключение к Telnet
  {
    client_telnet.flush();
    sprintf(str, "%d; %d; %d; %d; %d; %d; %d", serv_data.int_10V, serv_data.int_10dV, serv_data.int_h, serv_data.int_P1, serv_data.int_P2, serv_data.int_T1, serv_data.int_T2); 
    client_telnet.println(str);
  }

// HTTP     // пример тут: https://www.arduino.cc/en/Tutorial/WiFiChatServer
  client = http_server.available();
  if (client) // если подключился клиент
  {
    boolean currentLineIsBlank = true;
    while (client.connected())  // пока сохраняется подключение
      if (client.available())   // проверяем есть ли в буфере данные от клиентов (запросы)
      {
        char c = client.read(); // считываем символ из буфера
        if (c == '\n' && currentLineIsBlank) // ищем символ перехода строки, каждый ХТТП запрос им кончаетя, если его приняли то можно отправлять ответ
        {
          client.println("HTTP/1.1 200 OK\nContent-Type: text/html\nConnection: close\nRefresh: 1\n\n<!DOCTYPE HTML><html>");
          sprintf(str, "<h1>V= %d L (dm3)<br>dV= %d L (dm3)<br>h= %d mm<br>Pout= %d Pa<br>Pin= %d Pa<br>Tout= %d *C<br>Tin= %d *C</html>", serv_data.int_10V/10, serv_data.int_10dV/10, serv_data.int_h, serv_data.int_P1, serv_data.int_P2, serv_data.int_T1/10, serv_data.int_T2/10);
          client.println(str);
          break;
        }
        if (c == '\n') currentLineIsBlank = true;            // you're starting a new line
        else if (c != '\r') currentLineIsBlank = false;      // you've gotten a character on the current line
      }
    client.stop();  // закрываем соединение
  }
  else delay(100);  // без этой задержки не работает

// UDP
  char packetBuffer[sizeof(serv_data)];

  //if 
  (udp.parsePacket()) ;
  {
    udp.flush();
    
    memcpy(packetBuffer, &serv_data, sizeof(serv_data));
    
    IPAddress ip_milticast(0, 0, 0, 0);
    udp.beginPacket(ip_milticast, UDP_PORT);
    udp.write(packetBuffer, sizeof(serv_data));
    udp.endPacket();   
    delay(100); 
  }
}  


bool ConnectWiFi(const char *ssid, const char *pass) {
// 10 попыток соединения по WiFi
   for( int i=0; i<10; i++){
      Serial.print("\nConnecting to: "); 
      Serial.println(ssid); 
      WiFi.begin(ssid,pass);
      delay(1000);
// Максиммум 12 раз проверка соединения       
      for( int j=0; j<12; j++ ){
          if (WiFi.status() == WL_CONNECTED) return true; 
          delay(1000);
          Serial.print(WiFi.status()); 
      }
   }   
   Serial.println("\nConnect WiFi failed ...");
   return false;
} 








void do_led_ind()
{
  switch (Mode)
  {
   case IND_VOL : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_10V, 2, false); break;       // остаточный объем, литры с точностью 0,1
   case IND_DVOL : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_10dV, 2, false); break;     // расход, литры с точностью 0,1
   case IND_LEV : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_h, 2, false); break;         // текущий уровень, сантиметры с точностью 0,1
   case IND_P1 : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_P1, 0, false); break;         // Атмосферное давление, Па
   case IND_P2 : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_P2, 0, false); break;         // Давление на дне сосуда (атмосферное + гидростатическое), Па
   case IND_T1 : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_T1, 2, false); break;         // Температура платы измерителя, *C с точностью 0,1
   case IND_T2 : tm1638_disp.setDisplayToSignedDecNumber( serv_data.int_T2, 2, false); break;         // Температура жидкости, *C с точностью 0,1

   default : tm1638_disp.setDisplayToSignedDecNumber(serv_data.int_10V, 2, false); tm1638_disp.setLEDs(IND_VOL); break;             // остаточный объем, литры с точностью 0,1
  }
}

void readPressure(double *T1, double *T2, double *P1, double *P2)
{
  char temp;
    
  Wire.begin(BMP180_1_SDA, BMP180_1_SCL); // конфигурируем выводы и2ц интерфейса на пины, подкл. к датчику
  temp = bmp180_1.startTemperature();
  if (temp != 0) 
  {
    delay(temp);
    temp = bmp180_1.getTemperature(*T1);
    if (temp != 0)
    {
      temp = bmp180_1.startPressure(3);   //самое точное измерение
      if (temp != 0)
      {
        delay(temp);
        temp = bmp180_1.getPressure(*P1, *T1);
        
        if (temp == 0) Serial.println("Error read pressure 1");
       }
      else Serial.println("Error convert pressure 1");
    }
    else Serial.println("Error read temperature 1");
  } 
  else Serial.println("Error convert temperature 1");
  
  Wire.begin(BMP180_2_SDA, BMP180_2_SCL); // конфигурируем выводы и2ц интерфейса на пины, подкл. к датчику
  temp = bmp180_2.startTemperature();
  if (temp != 0) 
  {
    delay(temp);
    temp = bmp180_2.getTemperature(*T2);
    if (temp != 0)
    {
      temp = bmp180_2.startPressure(3);   //самое точное измерение
      if (temp != 0)
      {
        delay(temp);
        temp = bmp180_2.getPressure(*P2, *T2);
        
        if (temp == 0) Serial.println("Error read pressure 2");
       }
      else Serial.println("Error convert pressure 2");
    }
    else Serial.println("Error read temperature 2");
  } 
  else Serial.println("Error convert temperature 2");

}

